module RoboRace.Game where

import Control.Applicative ((<$>))
import Data.List
import Network.HTTP
import Text.ParserCombinators.Parsec

type GameId = Integer

-- | Identification of a game
data Game = Game
  { gameId :: GameId
  , gameServer :: String
  } deriving (Show)

-- | One state of a game
data GameState =
    NextTurn 
    { trField :: GameField -- Actual game field (does it change?)
    , trMoves :: [Move] -- Allowed moves
    , trPosition :: (Int,Int) -- Lerak's position
    , trRotation :: Rotation
    }
  | Won
    { trTotalMoves :: Int -- how many moves player needed to win
    }
  | Error
    { trErrorMsg :: String
    }
  deriving (Show)

-- | Field where Lerak moves
type GameField = [[FieldElem]]

-- | One element in a game field.
data FieldElem =
    Empty
  | Wall
  | Hole
  | Treasure

instance Show FieldElem where
  show Empty = "."
  show Wall = "X"
  show Hole = "#"
  show Treasure = "$"

data Move =
    GoForward Int
  | GoLeft
  | GoRight
  | GoBackward

data Rotation =
    North
  | West
  | South
  | East
  deriving (Show)

instance Show Move where
  show (GoForward 1) = "1"
  show (GoForward 2) = "2"
  show (GoForward 3) = "3"
  show GoLeft = "L"
  show GoRight = "R"
  show GoBackward = "Z"

charToMove :: Char -> Maybe Move
charToMove '1' = Just $ GoForward 1
charToMove '2' = Just $ GoForward 2
charToMove '3' = Just $ GoForward 3
charToMove 'L' = Just $ GoLeft
charToMove 'R' = Just $ GoRight
charToMove 'Z' = Just $ GoBackward
charToMove _  = Nothing

start :: String -> IO (Game,GameState)
start serverName = do
  let url = "http://" ++ serverName ++ "/start"
  res <- (simpleHTTP . getRequest $ url)
  resTxt <- getResponseBody res
  case parse gameState "(unknown)" resTxt of
    Left error ->
      fail $ show error
    Right (id,nextState) ->
      return 
        ( Game {
            gameId = id
          , gameServer = serverName
          }
        , nextState
        )
  
turn :: Game -> GameState -> [Move] -> IO GameState
turn game oldState moves = do
  let
    url = "http://" ++ gameServer game ++ "/kolo?id=" ++ (show $ gameId game) 
      ++ "&pohyby=" ++ concat (map show moves)
  res <- (simpleHTTP . getRequest $ url)
  resTxt <- getResponseBody res
  case parse gameState "(unknown)" resTxt of
    Left error ->
      fail $ show error
    Right (_,nextState) ->
      return nextState

-- Parser for game state read from server and game id.
gameState :: GenParser Char st (GameId,GameState)
gameState =
      (try wonMsg >>= return . (,) 0)
  <|> (try errorMsg >>= return . (,) 0)
  <|> nextTurnMsg

wonMsg :: GenParser Char st GameState
wonMsg = do
  string "Vyhrali jste za "
  moves <- read <$> many digit
  string " pohybu."
  return $ Won moves

errorMsg :: GenParser Char st GameState
errorMsg = do
  string "Chyba: "
  msg <- many anyChar
  return $ Error msg

nextTurnMsg :: GenParser Char st (GameId,GameState)
nextTurnMsg = do
  id <- gameIdLine
  field <- fieldLines
  pos <- position
  rot <- rotation
  moves <- allowedMoves
  return $ (id,NextTurn {
      trField = field
    , trMoves = moves
    , trPosition = pos
    , trRotation = rot
    })

gameIdLine :: GenParser Char st GameId
gameIdLine = do
  id <- read <$> many digit
  spaces
  return id

fieldLines :: GenParser Char st [[FieldElem]]
fieldLines =
  many fieldLine

fieldLine :: GenParser Char st [FieldElem]
fieldLine = do
  elems <- many fieldElem
  many1 space
  return elems

fieldElem :: GenParser Char st FieldElem
fieldElem =
  choice
    [ char '.' >> return Empty
    , char 'X' >> return Wall
    , char '#' >> return Hole
    , char '$' >> return Treasure
    ]

position :: GenParser Char st (Int,Int)
position = do
  spaces
  x <- read <$> many digit
  spaces
  y <- read <$> many digit
  return (x,y)

rotation :: GenParser Char st Rotation
rotation = do
  spaces
  choice
    [ char 'N' >> return North
    , char 'R' >> return East
    , char 'D' >> return South
    , char 'L' >> return West
    ]

allowedMoves :: GenParser Char st [Move]
allowedMoves = do
  spaces
  many1 allowedMove

allowedMove :: GenParser Char st Move
allowedMove = 
  choice 
    [ char '1' >> return (GoForward 1)
    , char '2' >> return (GoForward 2)
    , char '3' >> return (GoForward 3)
    , char 'L' >> return GoLeft
    , char 'R' >> return GoRight
    , char 'Z' >> return GoBackward
    ]
